#!/usr/bin/perl

use strict;
use warnings;
use GD;
use constant DEBUG => 0;
use Getopt::Long;

# constants {{{
use constant LEFT   => -1;
use constant TOP    => -1;
use constant CENTER => 0;
use constant BOTTOM => 1;
use constant RIGHT  => 1;

use constant JPEG => 0;
use constant PNG  => 1;
# /constants }}}

sub calc_text { # calculate text adjustment {{{
    my ($x, $y, $text, $font, $x_al, $y_al) = @_;

    die 'No font defined' unless defined($font);

    # _al - alignment
    # -1 left; 0 center; 1 right
    $x_al = CENTER unless defined $x_al;
    # -1 top; 0 center; 1 bottom
    $y_al = CENTER unless defined $y_al;

    my ($w, $h) = ($font->width*length($text), $font->height);
    # distance from (x, y) if not centered
    my $x_dist = 5;
    my $y_dist = 1;

    my ($tx, $ty) = ($x, $y);
    $tx = $x - $w        - $x_dist if $x_al == LEFT;
    $tx = $x - int($w/2)           if $x_al == CENTER;
    $tx = $x             + $x_dist if $x_al == RIGHT;

    $ty = $y - $h        - $y_dist if $y_al == TOP;
    $ty = $y - int($h/2)           if $y_al == CENTER;
    $ty = $y             + $y_dist if $y_al == BOTTOM;

    return ($tx, $ty, $w, $h);
} # }}}

sub dir { # select adjustment based on direction char {{{
    my ($dir) = @_;

    # default dir
    return (CENTER, CENTER) unless defined $dir;
    return (CENTER, CENTER) if $dir eq '';

    return (CENTER, CENTER) if $dir eq 'c';
    return (LEFT,   CENTER) if $dir eq 'l';
    return (RIGHT,  CENTER) if $dir eq 'r';
    return (CENTER, TOP)    if $dir eq 'a';
    return (CENTER, BOTTOM) if $dir eq 'u';
} # }}}

sub draw_grid { # draw a grid for easier orientation {{{
    my ($image) = @_;

    print 'Drawing grid ... ';
    my $step = 100;

    my $grid_col = $image->colorAllocateAlpha(127, 127, 127, 64);

    $image->setStyle($grid_col, $grid_col, gdTransparent, gdTransparent);

    # horiz
    for (my $x = 1; $x*$step <= $image->width; ++$x) {
        # normal dashed grid
        $image->line(
            $x*$step,              0,
            $x*$step, $image->height,
            gdStyled
        ) if $x % 5;
        # stroke every fifth line, triple stroke every tenth
        foreach ($x %  5, $x % 10, $x % 10) {
            $image->line(
                $x*$step,              0,
                $x*$step, $image->height,
                $grid_col,
            ) unless $_;
        }
        # grid rulers
        foreach (25, $image->height-25) {
            put_text(
                $image, $x * $step - int($step/2), $_,
                'H'. ($x-1), 5, 'g', 'c'
            );
        };
    }

    # vert
    for (my $y = 1; $y*$step <= $image->height; ++$y) {
        # normal dashed grid
        $image->line(
                        0, $y*$step,
            $image->width, $y*$step,
            gdStyled
        ) if $y % 5;
        # stroke every fifth line, triple stroke every tenth
        foreach ($y %  5, $y % 10, $y % 10) {
            $image->line(
                            0, $y*$step,
                $image->width, $y*$step,
                $grid_col,
            ) unless $_;
        }
        # grid rulers
        foreach (25, $image->width-25) {
            put_text(
                $image, $_, $y * $step - int($step/2),
                'V'. ($y-1), 5, 'g', 'c'
            );
        }
    }

    print "done\n";
} # }}}

sub mini_magic { # we don't need MIME::Magic {{{
    my ($testfile) = (@_);

    open(TEST, '<', $testfile) or die "Cannot open $testfile:\n$!";
    my $data;
    read (TEST, $data, 10);
    close(TEST);

    return JPEG if $data =~ m{JFIF}smo;
    return PNG  if $data =~ m{PNG}smo;

    return undef;
}; # }}}

STDOUT->autoflush(1);

# draw grid?
my $grid = 0;
# dither result image?
my $dither = 0;
# convert to palette image?
my $reduce = 0;
my $outfile;

GetOptions(
    'grid!' => \$grid,
    'dither!' => \$dither,
    'palette!' => \$reduce,
    'output=s' => \$outfile,
);

die 'Wrong number of arguments' if (@ARGV < 1);

# file name generation and checks {{{
# image (should be in JPEG/PNG format)
my $imgfile = $ARGV[0];
die "$imgfile not a file or not readable"
    unless (-f $imgfile) and (-r $imgfile);

# annotation text file (named like image but with extension txt)
my $txtfile = $imgfile;
$txtfile =~ s{\..*$}{.txt}o;
die "$txtfile not a file or not readable"
    unless (-f $txtfile) and (-r $txtfile);

# output file(named like image extension replaced by -new.png)
# add "-grid" and "-256" respectively
unless (defined $outfile) {
    $outfile = $imgfile;
    $outfile =~ s{\..*$}{-grid$&}o if $grid;
    $outfile =~ s{\..*$}{-256$&}o if $reduce;
    $outfile =~ s{\..*$}{-new.png}o;
}
# }}}

# detect type and load file {{{
my $type = mini_magic($imgfile);
GD::Image->trueColor(1);
my $image;
die "Unknown file type\n" unless defined($type);
print "$imgfile: ";
if ($type == JPEG) {
    print "JPEG file detected\n";
    $image = GD::Image->newFromJpeg($imgfile, 1);
}
if ($type == PNG) {
    print "PNG file detected\n";
    $image = GD::Image->newFromPng($imgfile, 1);
}
# work around truecolor bug in GD
# (will read 256 color PNG as palette even if truecolor is set)
unless ($image->isTrueColor) {
    print "Converting ". $image->colorsTotal ." color image to TrueColor\n";
    my $image_tmp = GD::Image->new($image->width, $image->height, 1);
    # copy image
    $image_tmp = $image->clone();
    $image = $image_tmp;
    undef $image_tmp;
}
die "Color mismatch" unless $image->isTrueColor;
# }}}

$image->alphaBlending(1);

my %color = ( # {{{
    # black   
    b => $image->colorAllocateAlpha(  0,   0,   0,  0),
    # red     
    R => $image->colorAllocateAlpha(255,   0,   0, 23),
    # green   
    G => $image->colorAllocateAlpha(  0, 255,   0, 23),
    # blue    
    B => $image->colorAllocateAlpha(  0,   0, 255, 23),
    # grey
    g => $image->colorAllocateAlpha(127, 127, 127,  0),
    # turquoise
    T => $image->colorAllocateAlpha(  0, 255, 255, 23),
    # magenta 
    M => $image->colorAllocateAlpha(255,   0, 255, 23),
    # yellow  
    Y => $image->colorAllocateAlpha(255, 255,   0, 23),
    # white   
    W => $image->colorAllocateAlpha(255, 255, 255,  0),
);
my %bgcolor = (
    # black   
    b => $image->colorAllocateAlpha(  0,   0,   0, 80),
    # red     
    R => $image->colorAllocateAlpha(255,   0,   0, 80),
    # green   
    G => $image->colorAllocateAlpha(  0, 255,   0, 80),
    # blue    
    B => $image->colorAllocateAlpha(  0,   0, 255, 80),
    # grey
    g => $image->colorAllocateAlpha(127, 127, 127, 80),
    # turquoise
    T => $image->colorAllocateAlpha(  0, 255, 255, 80),
    # magenta 
    M => $image->colorAllocateAlpha(255,   0, 255, 80),
    # yellow  
    Y => $image->colorAllocateAlpha(255, 255,   0, 80),
    # white   
    W => $image->colorAllocateAlpha(255, 255, 255, 80),
); # }}}

sub put_text { # put adjusted text/items on coords {{{
    my ($image, $x, $y, $text, $weight, $col, $dir) = @_;

    die 'Need text weight' unless defined $weight;
    #$col = 'W' unless defined $col;
    #$col = 'W' if $col eq '';
    $col = 'W' unless $col;

    # try to split off bg color
    my $bgcol;
    ($col, $bgcol) = $col =~ m{(.)(.?)}o;

    my $font;
    SELECT: {
        if ($weight <= 1) {
            # red dot if weight mode 0
            $image->filledRectangle($x-1,$y-1,$x+1,$y+1,$color{R})
                if $weight == 0;

            # with small text
            $font = gdTinyFont;
            last SELECT;
        }
        if ($weight == 2) {
            # medium text
            $font = gdSmallFont;
            last SELECT;
        }
        if ($weight == 3) {
            # large text
            $font = gdMediumBoldFont;
            last SELECT;
        }
        if ($weight == 4) {
            # large text
            $font = gdLargeFont;
            last SELECT;
        }
        if ($weight == 5) {
            # large text
            $font = gdGiantFont;
            last SELECT;
        }
        if ($weight == 8) {
            # special case to draw regular rectangles
            my ($endx, $endy) =
                $text =~ m{^\s* ([0-9]+) ,\s* ([0-9]+) \s*$}ox;
            $image->rectangle($x, $y, $endx, $endy, $color{$col});
            return;
        }
        if ($weight == 9) {
            # special case to draw lines
            my ($endx, $endy) =
                $text =~ m{^\s* ([0-9]+) ,\s* ([0-9]+) \s*$}ox;
            $image->line($x, $y, $endx, $endy, $color{$col});
            return;
        }
    }
    my ($tx, $ty, $tw, $th) = calc_text($x, $y, $text, $font, dir($dir));

    # draw background
    $image->filledRectangle($tx-1, $ty,$tx-1 + $tw, $ty + $th, $bgcolor{$bgcol})
        if $bgcol;
    # put text
    $image->string($font, $tx, $ty, $text, $color{$col});
} # }}}

draw_grid($image) if $grid;

open(ANN, '<', $txtfile)
    or die "Cannot read annotations from $txtfile:\n$!";
print 'Annotating image ';
DO_ANNOTATE: while(<ANN>) {
    chomp;

    # silently ignore comments and empty lines
    next DO_ANNOTATE if m{^(#|$)}o;
    # remove comments after lines
    s{\s+#\s+.*$}{}o;

    # split values {{{
    my ($x, $y, $weight, $dir, $color, $name) = m{
        # coords
        ^\s*([0-9]+), \s*([0-9]+), \s*
        # weight
        ([0-589])
        # direction
        ([aclru]?)
        # color and optional background color
        ([bBgGMRTWY]{0,2})\s+
        # location's name
        (.*)$
    }ox; # }}}

    # some checks {{{
    die "\nError in line '$_'" unless defined $name;
    if (($x<0) or ($x>$image->width)) {
        print "Warning: out of bounds -- *$x*, $y\n";
        $x = 0             if $x < 0;
        $x = $image->width if $x > $image->width;
    }
    if (($y<0) or ($y>$image->height)) {
        print "Warning: out of bounds -- $x, *$y*\n";
        $y = 0              if $y < 0;
        $y = $image->height if $y > $image->height;
    } # }}}

    if (DEBUG) { # draw red cross through (x, y) {{{
        my $rad = 10;
        $image->line($x-$rad, $y, $x+$rad, $y, $color{R});
        $image->line($x, $y-$rad, $x, $y+$rad, $color{R});
    } # }}}

    put_text($image, $x, $y, $name, $weight, $color, $dir) unless DEBUG > 1;

    print '.' unless DEBUG;
}
close(ANN);
print " done\n";

if ($reduce) { # convert to palette image # {{{
    print 'Reducing colors ... ';
    # reduce colors
    $image->trueColorToPalette(($dither ? 1 : 0), 256);
    print "done\n";
} # }}}

print "Writing image $outfile ... "; # {{{
my $png_data = $image->png(9); # best compression
open (OUT, '>', $outfile) or
    die "Cannot write output file $outfile:\n$!";
binmode OUT;
print OUT $png_data;
close OUT;
print "done\n"; # }}}

exit 0;
# vim: foldmethod=marker
