#!/usr/bin/perl

use strict;
use warnings;
use GD;
use Getopt::Long;

# constants {{{
use constant LEFT   => -1;
use constant TOP    => -1;
use constant CENTER => 0;
use constant BOTTOM => 1;
use constant RIGHT  => 1;

use constant JPEG => 0;
use constant PNG  => 1;
# /constants }}}

sub mini_magic { # we don't need MIME::Magic {{{
    my ($testfile) = (@_);

    open(TEST, '<', $testfile) or die "Cannot open $testfile:\n$!";
    my $data;
    read (TEST, $data, 10);
    close(TEST);

    return JPEG if $data =~ m{JFIF}smo;
    return PNG  if $data =~ m{PNG}smo;

    return undef;
}; # }}}

STDOUT->autoflush(1);

# dither result image?
my $dither = 0;
# convert to palette image?
my $reduce = 0;
my $outbase;

GetOptions(
    'dither!' => \$dither,
    'palette!' => \$reduce,
    'output_base|outbase=s' => \$outbase,
);

die 'Wrong number of arguments' if (@ARGV < 1);

# file name generation and checks {{{
# image (should be in JPEG/PNG format)
my $imgfile = $ARGV[0];
die "$imgfile not a file or not readable"
    unless (-f $imgfile) and (-r $imgfile);

# cut coordinate file (named like image but with extension cut)
my $cutfile = $imgfile;
$cutfile =~ s{\..*$}{.cut}o;
die "$cutfile not a file or not readable"
    unless (-f $cutfile) and (-r $cutfile);

# output file(named like image extension replaced by -new.png)
unless (defined $outbase) {
    $outbase = $imgfile;
    # remove extension
    $outbase =~ s{\..*$}{}o;
    # add "-256" respectively
    $outbase .= '-256' if $reduce;
}
# }}}

# detect type and load file {{{
my $type = mini_magic($imgfile);
GD::Image->trueColor(1);
my $image;
die "Unknown file type\n" unless defined($type);
print "$imgfile: ";
if ($type == JPEG) {
    print "JPEG file detected\n";
    $image = GD::Image->newFromJpeg($imgfile, 1);
}
if ($type == PNG) {
    print "PNG file detected\n";
    $image = GD::Image->newFromPng($imgfile, 1);
}
## work around truecolor bug in GD
## (will read 256 color PNG as palette even if truecolor is set)
#unless ($image->isTrueColor) {
#    print "Converting ". $image->colorsTotal ." color image to TrueColor\n";
#    my $image_tmp = GD::Image->new($image->width, $image->height, 1);
#    # copy image
#    $image_tmp = $image->clone();
#    $image = $image_tmp;
#    undef $image_tmp;
#}
#die "Color mismatch" unless $image->isTrueColor;
# }}}

$image->alphaBlending(1);

open(CUT, '<', $cutfile)
    or die "Cannot read cut coordinates from $cutfile:\n$!";
DO_CUT: while(<CUT>) {
    chomp;

    # silently ignore comments and empty lines
    next DO_CUT if m{^(#|$)}o;
    # remove comments after lines
    s{\s+#\s+.*$}{}o;

    # split values {{{
    my ($x, $y, $width, $height, $name) = m{
        # coords start
        ^\s*([0-9]+), \s*([0-9]+), \s*
        # coords end
        \s*([0-9]+), \s*([0-9]+) \s+
        # segment name
        (.*)$
    }ox; # }}}

    printf "Cutting out: (%d, %d) - (%d, %d) '%s'\n",
        $x, $y, $x+$width, $y+$height, $name;

    # some checks {{{
    die "\nError in line '$_'" unless defined $name;
    if (($x<0) or ($x>$image->width)) {
        print "Warning: out of bounds -- x: $x\n";
        next DO_CUT;
    }
    if (($y<0) or ($y>$image->height)) {
        print "Warning: out of bounds -- y: $y\n";
        next DO_CUT;
    }
    if (($x+$width<0) or ($x+$width>$image->width)) {
        print 'Warning: out of bounds -- width: '. $x+$width ."\n";
        next DO_CUT;
    }
    if (($y+$height<0) or ($y+$height>$image->height)) {
        print 'Warning: out of bounds -- height: '. $y+$height ."\n";
        next DO_CUT;
    }
    # }}}

    # finally cut
    my $segment = GD::Image->new($width, $height, 1);
    $segment->copy($image, 0, 0 , $x, $y, $width, $height);

    if ($reduce) { # convert to palette image # {{{
        print 'Reducing colors ... ';
        # reduce colors
        $segment->trueColorToPalette(($dither ? 1 : 0), 256);
        print "done\n";
    } # }}}

    my $outfile = $outbase .'-'. $name .'.png';
    print "Writing image $outfile ... "; # {{{
    my $png_data = $segment->png(9); # best compression
    undef $segment;
    open (OUT, '>', $outfile) or
        die "Cannot write output file $outfile:\n$!";
    binmode OUT;
    print OUT $png_data;
    close OUT;
    print "done\n"; # }}}

}
close(CUT);
print "All done\n";

exit 0;
# vim: foldmethod=marker
